import logging

FORMAT = "%(asctime)s [%(filename)s] [%(levelname)s] - %(message)s"
LEVEL = logging.INFO

logging.basicConfig(format=FORMAT, level=LEVEL)


def get_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    return logger



from enum import Enum


class OrderSelectionMethodAR(Enum):
    PACF = "pacf"

    def __str__(self) -> str:
        return self._value_


class FitMethodAR(Enum):
    YULE_WALKER = ("yule_walker", False)
    BURG = ("burg", False)
    CMLE_WITHOUT_CST = ("cmle_without_cst", False)
    CMLE_WITH_CST = ("cmle_with_cst", True)
    MLE_WITHOUT_CST = ("mle_without_cst", False)
    MLE_WITH_CST = ("mle_with_cst", True)

    def __new__(cls, name: str, has_cst_parameter: bool):
        obj = object.__new__(cls)
        obj._value_ = name
        obj.has_cst_parameter = has_cst_parameter
        return obj

    def __str__(self) -> str:
        return self._value_



class IllegalNbLagsException(Exception):
    """Raised when the number of lags for a given time series is not valid"""
    pass


class IllegalValueException(Exception):
    """Raised when a value is not in a valid state"""
    pass


class ModelNotFittedException(Exception):
    """Raised when the model need the parameters, but it has not been fitted"""
    pass


class ModelNotSimulatedException(Exception):
    """Raised when the model need the simulated time series, but it has not been simulated"""
    pass


class NonStationaryTimeSeriesException(Exception):
    """Raised when the time series is not stationary"""
    pass


class AutocorrelatedResidualsException(Exception):
    """Raised when the residuals are autocorrelated, based on the result of a hypothesis test"""
    pass


class NoConvergenceException(Exception):
    """Raised when the method used to estimate model parameters failed to converge"""
    pass



from __future__ import annotations

import ctypes as ct
import platform
from pathlib import Path
from typing import Dict, Tuple, Union, List

from tradeflow.common import logger_utils
from tradeflow.common.config import LIBTRADEFLOW_SHARED_LIBRARY_DIRECTORY
from tradeflow.common.exceptions import UnsupportedOsException, SharedLibraryNotFoundException
from tradeflow.common.general_utils import check_condition
from tradeflow.common.singleton import Singleton

logger = logger_utils.get_logger(__name__)


class SharedLibrariesRegistry(metaclass=Singleton):

    def __init__(self) -> None:
        self._name_to_shared_library: Dict[str, SharedLibrary] = {}
        self._set_shared_libraries(self._get_shared_libraries())

    def _set_shared_libraries(self, shared_libraries: List[SharedLibrary]) -> None:
        for shared_library in shared_libraries:
            self._add_shared_library(shared_library=shared_library)

    def _get_shared_libraries(self) -> List[SharedLibrary]:
        # simulate: size (int), inverted_params (double*), constant_parameter (double), nb_params (int), last_signs (int*), seed (int), res (int*)
        simulate = Function(name="simulate", argtypes=[ct.c_int, ct.POINTER(ct.c_double), ct.c_double, ct.c_int, ct.POINTER(ct.c_int), ct.c_int, ct.POINTER(ct.c_int)], restype=ct.c_void_p)
        libtradeflow = SharedLibrary(name="libtradeflow", directory=LIBTRADEFLOW_SHARED_LIBRARY_DIRECTORY, functions=[simulate])
        return [libtradeflow]

    def _add_shared_library(self, shared_library: SharedLibrary) -> SharedLibrariesRegistry:
        assert shared_library.name not in self._name_to_shared_library
        self._name_to_shared_library[shared_library.name] = shared_library
        logger.info(f"Added shared library '{shared_library.name}' to the registry")
        return self

    def find_shared_library(self, name: str) -> SharedLibrary:
        check_condition(condition=name in self._name_to_shared_library, exception=SharedLibraryNotFoundException(f"Shared library '{name}' not found in the registry."))
        return self._name_to_shared_library.get(name)


class SharedLibrary:

    LINUX = "linux"
    DARWIN = "darwin"
    WINDOWS = "windows"

    SO = "so"
    DYLIB = "dylib"
    DLL = "dll"

    ARGUMENT_TYPES = "argtypes"
    RESULT_TYPE = "restype"

    def __init__(self, name: str, directory: Path, functions: List[Function]) -> None:
        self._name = name
        self._directory = directory
        self._functions = functions
        self._cdll = None  # Loaded shared library is cached after the first call to load()

    @property
    def name(self):
        return self._name

    def load(self) -> ct.CDLL:
        """
        Return the shared library of the project.

        Returns
        -------
        ct.CDLL
            The loaded shared library.
        """
        if self._cdll is not None:
            return self._cdll

        shared_library_extension = SharedLibrary.get_shared_library_extension()
        shared_library_path = self._directory.joinpath(f"{self._name}.{shared_library_extension}")
        check_condition(condition=shared_library_path.exists() and shared_library_path.is_file(),
                        exception=FileNotFoundError(f"Shared library '{self._name}.{shared_library_extension}' not found in directory '{self._directory}'."))

        cdll = ct.CDLL(str(shared_library_path), winmode=0)
        for function in self._functions:
            setattr(getattr(cdll, function.name), self.ARGUMENT_TYPES, function.argtypes)
            setattr(getattr(cdll, function.name), self.RESULT_TYPE, function.restype)

        logger.info(f"Loaded shared library '{self._name}'")
        self._cdll = cdll
        return self._cdll

    @staticmethod
    def get_shared_library_extension() -> str:
        """
        Determine the shared library file extension based on the operating system.

        Returns
        -------
        str
            The file extension for shared libraries, specific to the current operating system.

        Raises
        ------
        UnsupportedOsException
            If the operating system is not Linux, Darwin (macOS), or Windows.
        """
        os_name_to_shared_library_extension = {
            SharedLibrary.LINUX: SharedLibrary.SO,
            SharedLibrary.DARWIN: SharedLibrary.DYLIB,
            SharedLibrary.WINDOWS: SharedLibrary.DLL,
        }

        os_name = platform.system().lower()
        extension = os_name_to_shared_library_extension.get(os_name)

        check_condition(condition=extension is not None, exception=UnsupportedOsException(f"Unsupported OS '{os_name}'. Supported OS values are Linux, Darwin, and Windows."))
        return extension


class Function:

    def __init__(self, name: str, argtypes: List[Union[ct._CData, ct.POINTER(ct._CData)]], restype: Union[ct._CData, ct.POINTER(ct._CData)]) -> None:
        self._name = name
        self._argtypes = tuple(argtypes)
        self._restype = restype

    @property
    def name(self) -> str:
        return self._name

    @property
    def argtypes(self) -> Tuple[Union[ct._CData, ct.POINTER(ct._CData)], ...]:
        return self._argtypes

    @property
    def restype(self) -> Union[ct._CData, ct.POINTER(ct._CData)]:
        return self._restype



from threading import Lock


class Singleton(type):

    _instances = {}
    _lock = Lock()

    def __call__(cls, *args, **kwargs):
        with cls._lock:
            if cls not in cls._instances:
                cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]



import ctypes as ct
from typing import Literal, Any

from statsmodels.tools.typing import ArrayLike1D

from tradeflow.common import logger_utils

logger = logger_utils.get_logger(__name__)


class CArray:

    @staticmethod
    def of(c_type_str: Literal["int", "double"], arr: ArrayLike1D) -> ct.Array:
        """
        Create a ctypes array from a Python list.

        Parameters
        ----------
        c_type_str : {'int', 'double'}
            The type of the array to be created.
        arr : array_like
            The array from which to create the ctypes array.

        Returns
        -------
        ct.Array
            The ctypes array containing the elements of `arr`.
        """
        c_type = get_c_type_from_string(c_type_str=c_type_str)
        return (c_type * len(arr))(*arr)


class CArrayEmpty:

    @staticmethod
    def of(c_type_str: Literal["int", "double"], size: int) -> ct.Array:
        """
        Create an empty ctypes array of a given size.

        Parameters
        ----------
        c_type_str : {'int', 'double'}
            The type of elements in the array to be created.
        size : int
            The size of the ctypes array to create.

        Returns
        -------
        ct.Array
            The empty ctypes array of size `size`.
        """
        c_type_str = get_c_type_from_string(c_type_str=c_type_str)
        return (c_type_str * size)()


def get_c_type_from_string(c_type_str: Literal["int", "double"]) -> Any:
    """
    Return a ctypes type corresponding to a given C data type (from a string).

    Parameters:
    -----------
    c_type_str : {'int', 'double'}
        A string indicating the desired C data type.

    Returns:
    --------
    ct._SimpleCData
        The corresponding ctypes type.
    """
    c_type_str_to_c_type = {
        "int": ct.c_int,
        "double": ct.c_double
    }

    if c_type_str not in c_type_str_to_c_type:
        raise Exception(f"Unknown type {c_type_str}")

    return c_type_str_to_c_type[c_type_str]



from __future__ import annotations

from typing import Literal, Optional

import numpy as np
import scipy.optimize as optimize
from numpy.linalg import slogdet
from statsmodels.regression import yule_walker
from statsmodels.regression.linear_model import burg, OLS
from statsmodels.tools import add_constant
from statsmodels.tools.typing import ArrayLike1D
from statsmodels.tsa.ar_model import sumofsq
from statsmodels.tsa.tsatools import lagmat

from tradeflow.common import logger_utils
from tradeflow.common.ctypes_utils import CArray, CArrayEmpty
from tradeflow.common.general_utils import check_condition, check_enum_value_is_valid, get_enum_values, \
    is_value_within_interval_exclusive
from tradeflow.common.shared_libraries_registry import SharedLibrariesRegistry
from tradeflow.config import LIB_TRADEFLOW
from tradeflow.enums import OrderSelectionMethodAR, FitMethodAR
from tradeflow.exceptions import IllegalValueException, ModelNotFittedException, IllegalNbLagsException, \
    NonStationaryTimeSeriesException, AutocorrelatedResidualsException, NoConvergenceException
from tradeflow.time_series import TimeSeries

logger = logger_utils.get_logger(__name__)
LOG_2_PI = np.log(2 * np.pi)


class AR(TimeSeries):
    """
    Autoregressive model for trade/order signs.

    Parameters
    ----------
    signs : array_like
        An array of signs where each element is either 1 (representing a buy) or -1 (representing a sell).
    max_order : int, default None
        The maximum order of the autoregressive model.
        If None, the maximum order is set to 12*(nobs/100)^{1/4} as outlined in Schwert (1989).
    order_selection_method : {'pacf'}, default None
        The method for selecting the order of the model. If None, the order of the model will be `max_order`.

        * 'pacf' - Choose the model using the number of significant lags in the partial autocorrelation function of the time series of signs.
    """

    def __init__(self, signs: ArrayLike1D, max_order: Optional[int] = None,
                 order_selection_method: Optional[Literal["pacf"]] = None) -> None:
        super().__init__(signs=signs)
        self._max_order = self._init_max_order(max_order=max_order)
        self._order_selection_method = check_enum_value_is_valid(enum_obj=OrderSelectionMethodAR,
                                                                 value=order_selection_method,
                                                                 parameter_name="order_selection_method",
                                                                 is_none_valid=True)

        # Will be set during fit()
        self._constant_parameter = 0.0
        self._parameters = None

    def resid(self, seed: Optional[int] = None) -> np.ndarray:
        """
        Estimate and return the residuals of the model.

        Residuals are calculated as the difference between the observed values and the
        values predicted by the model using the fitted parameters.

        Parameters
        ----------
        seed : int, default None
            Seed used to initialize the pseudo-random number generator.
            If `seed` is `None`, then a random seed between 0 and 2^32 - 1 is used.

        Returns
        -------
        np.ndarray
            The residuals of the AR model.

        Raises
        ------
        ModelNotFittedException
            If the model's parameters are not set. This occurs if the model
            has not been fitted by calling `fit()`.

        """
        check_condition(condition=self._parameters is not None, exception=ModelNotFittedException("The model does not have its parameters set. Fit the model first by calling 'fit()'."))

        if seed is None:
            seed = np.random.randint(0, np.iinfo(np.int32).max)
        np.random.seed(seed)

        def predict_sign_from_expected_value(expected_value: float, uniform: float) -> int:
            buy_proba = 0.5 * (1 + expected_value)
            next_sign = 1 if uniform <= buy_proba else -1
            return next_sign

        x, y = lagmat(x=self._signs, maxlag=len(self._parameters), trim="both", original="sep", use_pandas=False)
        x_with_cst = np.c_[np.ones(shape=x.shape[0]), x]

        expected_values_pred = x_with_cst @ np.append(self._constant_parameter, self._parameters)
        uniforms = np.random.uniform(low=0, high=1, size=len(expected_values_pred))
        signs_pred = np.vectorize(predict_sign_from_expected_value)(expected_values_pred, uniforms)

        resid = y.squeeze() - signs_pred
        return resid

    def _init_max_order(self, max_order: Optional[int]) -> int:
        if max_order is None:
            # Schwert (1989)
            max_order = int(np.ceil(12.0 * np.power(self._nb_signs / 100.0, 1 / 4.0)))

        check_condition(condition=1 <= max_order < self._nb_signs // 2,
                        exception=IllegalNbLagsException(f"{max_order} is not valid for 'max_order', it must be positive and lower than 50% of the time series length (< {self._nb_signs // 2})."))
        logger.info(f"The maximum order has been set to {max_order}.")
        return max_order

    def fit(self, method: Literal["yule_walker", "burg", "cmle_without_cst", "cmle_with_cst", "mle_without_cst", "mle_with_cst"], significance_level: float = 0.05, check_stationarity: bool = True, check_residuals: bool = True) -> AR:
        """
        Estimate the model parameters.

        If the chosen method estimates a constant term, the proportion of buy signs in the time series generated with
        these parameters will be close to the one from the training time series ('cmle_with_cst', 'mle_with_cst').

        Otherwise, the proportion of buy signs in the time series generated with these parameters
        will be close to 50% ('yule_walker', 'burg', 'cmle_without_cst', 'mle_without_cst').

        Parameters
        ----------
        method : {'yule_walker', 'burg', 'cmle_without_cst', 'cmle_with_cst', 'mle_without_cst', 'mle_with_cst'}
            The method to use for estimating parameters.

            * 'yule_walker' - Use the Yule-Walker equations to estimate model parameters.
            * 'burg' - Use Burg's method to estimate model parameters.
            * 'cmle_without_cst' - Use conditional maximum likelihood estimation without constant term to estimate model parameters.
              It can be solved with an OLS.
            * 'cmle_with_cst' - Use conditional maximum likelihood estimation with a constant term to estimate model parameters.
              It can be solved with an OLS.
            * 'mle_without_cst' - Use maximum likelihood estimation without constant term to estimate model parameters.
            * 'mle_with_cst' - Use maximum likelihood estimation with a constant term to estimate model parameters.
        significance_level : float, default 0.05
            The significance level for stationarity (if `check_stationarity` is `True`) and residual autocorrelation (if `check_residuals` is `True`) tests.
        check_stationarity : bool, default True
            If `True`, performs a stationarity check on the time series using the Augmented Dickey-Fuller unit root test.
            Raises an exception if the time series is not stationary according to the test.
        check_residuals : bool, default True
            If `True`, performs a residual autocorrelation check using the Breusch-Godfrey test.
            Raises an exception if residuals are autocorrelated.

        Returns
        -------
        AR
            The AR instance.
        """
        method: FitMethodAR = check_enum_value_is_valid(enum_obj=FitMethodAR, value=method, parameter_name="method", is_none_valid=False)
        self._select_order()

        if check_stationarity:
            check_condition(
                condition=self.is_time_series_stationary(time_series=self._signs, nb_lags=self._order, significance_level=significance_level, regression="n"),
                exception=NonStationaryTimeSeriesException("The time series must be stationary in order to be fitted. You can set 'check_stationarity' to False to disable this check.")
            )

        parameters = None
        if method == FitMethodAR.YULE_WALKER:
            parameters = yule_walker(x=self._signs, order=self._order, method="mle", df=None, inv=False, demean=True)[0]
        elif method == FitMethodAR.BURG:
            parameters, _ = burg(endog=self._signs, order=self._order, demean=True)
        elif method in (FitMethodAR.CMLE_WITHOUT_CST, FitMethodAR.CMLE_WITH_CST):
            self._x, self._y = self._get_model_x_y(has_cst_parameter=method.has_cst_parameter)
            ols = OLS(endog=self._y, exog=self._x, missing="raise", hasconst=method.has_cst_parameter).fit()
            parameters = ols.params
        elif method in (FitMethodAR.MLE_WITHOUT_CST, FitMethodAR.MLE_WITH_CST):
            self._x, self._y = self._get_model_x_y(has_cst_parameter=method.has_cst_parameter)
            self._start_idx_parameters = 1 if method.has_cst_parameter else 0
            self._first_order_signs = self._signs[:self._order].reshape((self._order, 1))

            def f(params: np.ndarray) -> float:
                return -self._log_likelihood(parameters=params) / self._nb_signs

            start_parameters = self._compute_start_parameters(has_cst_parameter=method.has_cst_parameter)
            parameters, _, res = optimize.fmin_l_bfgs_b(func=f, x0=start_parameters, approx_grad=True, factr=1e2, pgtol=1e-8)

            if res["warnflag"] != 0:
                raise NoConvergenceException("lbfgs method failed to find optimal parameters, you may try to use another method.")

            logger.info(f"Found optimal parameters for MLE using lbfgs in {res['nit']} iterations.")
        else:
            raise IllegalValueException(
                f"The method '{method}' for the parameters estimation is not valid, it must be among {get_enum_values(enum_obj=FitMethodAR)}.")

        self._set_parameters(parameters=parameters, has_cst_parameter=method.has_cst_parameter)

        if check_residuals:
            _, p_value = self.breusch_godfrey_test(resid=self.resid(seed=1))
            # If the p value is below the significance level, we can reject the null hypothesis of no autocorrelation
            logger.info(f"Breusch-Godfrey test: p value for the null hypothesis of no autocorrelation is {round(p_value, 4)}.")
            check_condition(condition=p_value > significance_level,
                            exception=AutocorrelatedResidualsException(f"The residuals of the model seems to be autocorrelated (p value of the null hypothesis of no autocorrelation is {round(p_value, 4)}), "
                                                                       f"you may try to increase the number of lags, or you can set 'check_residuals' to False to disable this check."))

        logger.info(f"The AR({self._order}) model has been fitted with method '{method}'.")
        return self

    def _get_model_x_y(self, has_cst_parameter: bool) -> tuple[np.ndarray, np.ndarray]:
        x, y = lagmat(x=self._signs, maxlag=self._order, trim="both", original="sep", use_pandas=False)
        if has_cst_parameter:
            x = add_constant(data=x, prepend=True, has_constant="raise")
        return x, y

    def _compute_start_parameters(self, has_cst_parameter: bool) -> np.ndarray:
        if has_cst_parameter:
            ols = OLS(endog=self._y, exog=self._x, missing="raise", hasconst=has_cst_parameter).fit()
            return ols.params
        else:
            return yule_walker(x=self._signs, order=self._order, method="mle", df=None, inv=False, demean=True)[0]

    def _select_order(self) -> None:
        if self._order_selection_method is None:
            self._order = self._max_order
        elif self._order_selection_method == OrderSelectionMethodAR.PACF:
            pacf_coeffs, confidence_interval = self.calculate_pacf(nb_lags=self._max_order, alpha=0.05)

            pacf_coeffs = pacf_coeffs[1:]
            confidence_interval = confidence_interval[1:]

            lower_band = confidence_interval[:, 0] - pacf_coeffs
            upper_band = confidence_interval[:, 1] - pacf_coeffs

            order = 0
            for acf_coeff, value_lower_band, value_upper_band in zip(pacf_coeffs, lower_band, upper_band):
                if is_value_within_interval_exclusive(value=acf_coeff, lower_bound=value_lower_band, upper_bound=value_upper_band):
                    break
                order += 1
            self._order = order
        else:
            raise IllegalValueException(
                f"The method '{self._order_selection_method}' for the order selection is not valid, it must be among {get_enum_values(enum_obj=OrderSelectionMethodAR)}")

        logger.info(f"AR order selection: {self._order} lags (method: {self._order_selection_method}, time series length: {self._nb_signs}).")

    def _log_likelihood(self, parameters: np.ndarray) -> float:
        # Time Series Analysis - Hamilton, J.D, (5.3.6, p. 124).
        constant_parameter = 0
        if self._start_idx_parameters != 0:
            constant_parameter = parameters[0]

        # Vector filled with the mean value
        mu = constant_parameter / (1 - np.sum(parameters[self._start_idx_parameters:]))
        mu_p = np.full(shape=(self._order, 1), fill_value=mu, dtype=float)

        # Difference between first order observations and mu
        diff_p = self._first_order_signs - mu_p

        vp_inv = self._calculate_vp_inv(parameters=parameters.copy())
        diff_p_vp_inv = np.dot((np.dot(diff_p.T, vp_inv)), diff_p).item()

        pred = np.dot(self._x, parameters)
        sum_square_residuals = sumofsq(self._y.squeeze() - pred)
        sigma2 = 1.0 / self._nb_signs * (diff_p_vp_inv + sum_square_residuals)

        log_determinant_vp_inv = slogdet(vp_inv)[1]
        log_likelihood = -0.5 * (self._nb_signs * (LOG_2_PI + np.log(sigma2)) - log_determinant_vp_inv + diff_p_vp_inv / sigma2 + sum_square_residuals / sigma2)
        return log_likelihood

    def _calculate_vp_inv(self, parameters: np.ndarray) -> np.ndarray:
        # Time Series Analysis - Hamilton, J.D, (5.3.7, p. 125).
        parameters = np.r_[-1, parameters[self._start_idx_parameters:]]
        vp_inv = np.zeros(shape=(self._order, self._order), dtype=float)

        for i in range(1, self._order + 1):
            vp_inv[i - 1, i - 1:] = np.correlate(a=parameters, v=parameters[:i])[:-1] - np.correlate(a=parameters[-i:], v=parameters)[:-1]

        vp_inv = vp_inv + vp_inv.T - np.diag(vp_inv.diagonal())
        return vp_inv

    def _set_parameters(self, parameters: np.ndarray, has_cst_parameter: bool) -> None:
        if parameters is None:
            raise Exception("Can't set parameters to None.")

        if has_cst_parameter:
            if len(parameters) != self._order + 1:
                raise Exception(f"Expected {self._order + 1} parameters (including constant term), but got {len(parameters)}.")
            self._constant_parameter, self._parameters = parameters[0], parameters[1:]
        else:
            if len(parameters) != self._order:
                raise Exception(f"Expected {self._order} parameters, but got {len(parameters)}.")
            self._parameters = parameters

    def simulate(self, size: int, seed: Optional[int] = None) -> np.ndarray:
        """
        Simulate a time series of signs after the model has been fitted.

        Parameters
        ----------
        size : int
            The number of signs to simulate.
        seed : int, default None
            Seed used to initialize the pseudo-random number generator.
            If `seed` is `None`, then a random seed between 0 and 2^32 - 1 is used.

        Returns
        -------
        np.ndarray
            The simulated signs (+1 for buy, -1 for sell).
        """
        check_condition(condition=size > 0, exception=IllegalValueException(f"The size '{size}' for the time series to be simulated is not valid, it must be greater than 0."))
        check_condition(condition=self._parameters is not None, exception=ModelNotFittedException("The model has not yet been fitted. Fit the model first by calling 'fit()'."))

        if seed is None:
            seed = np.random.randint(0, np.iinfo(np.int32).max)

        inverted_parameters = CArray.of(c_type_str="double", arr=self._parameters[::-1])
        last_signs = CArray.of(c_type_str="int", arr=np.asarray(self._signs[-self._order:]).astype(int))
        self._simulation = CArrayEmpty.of(c_type_str="int", size=size)

        cpp_lib = SharedLibrariesRegistry().find_shared_library(name=LIB_TRADEFLOW).load()
        cpp_lib.simulate(size, inverted_parameters, self._constant_parameter, len(inverted_parameters), last_signs, seed, self._simulation)
        return self._simulation[:]



import os
import re
import subprocess
import sys
from pathlib import Path

import toml
from setuptools import Extension, setup, find_packages
from setuptools.command.build_ext import build_ext

PACKAGE_NAME = toml.load(Path(__file__).parent.joinpath("pyproject.toml"))["project"]["name"]


class CMakeExtension(Extension):

    def __init__(self, name: str):
        super().__init__(name, sources=[])


class CMakeBuild(build_ext):

    def build_extension(self, ext: CMakeExtension) -> None:
        cwd = Path().cwd()
        extdir = cwd.joinpath(self.get_ext_fullpath(ext.name)).parent.resolve()

        debug = int(os.environ.get("DEBUG", 0)) if self.debug is None else self.debug
        build_type = "Debug" if debug else "Release"

        cmake_args = [
            f"-DMY_PROJECT_SOURCE_DIR={extdir}",
            f"-DCMAKE_BUILD_TYPE={build_type}"
        ]
        build_args = ["--config", build_type]

        if sys.platform.startswith("darwin"):
            # Cross-compile support for macOS - respect ARCHFLAGS if set
            archs = re.findall(r"-arch (\S+)", os.environ.get("ARCHFLAGS", ""))
            if archs:
                cmake_args += ["-DCMAKE_OSX_ARCHITECTURES={}".format(";".join(archs))]

        build_temp = Path(self.build_temp).joinpath(ext.name)
        build_temp.mkdir(parents=True, exist_ok=True)

        subprocess.run(
            ["cmake", cwd, *cmake_args], cwd=build_temp, check=True
        )
        subprocess.run(
            ["cmake", "--build", ".", *build_args], cwd=build_temp, check=True
        )


setup(
    packages=find_packages(exclude=["scripts"]),
    ext_modules=[CMakeExtension(name=PACKAGE_NAME)],
    cmdclass={'build_ext': CMakeBuild}
)



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Literal, Tuple, Any, Optional, List

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import statsmodels.tsa.stattools as stattools
from matplotlib.figure import Figure
from scipy import stats
from statsmodels.regression.linear_model import OLS
from statsmodels.tools.typing import ArrayLike1D
from statsmodels.tools.validation import bool_like
from statsmodels.tsa.stattools import acf, pacf
from statsmodels.tsa.tsatools import lagmat

from tradeflow.common import logger_utils
from tradeflow.common.general_utils import check_condition
from tradeflow.exceptions import IllegalNbLagsException, IllegalValueException, \
    ModelNotSimulatedException

logger = logger_utils.get_logger(__name__)


class TimeSeries(ABC):
    """
    Time series model for trade/order signs. Intended to be subclassed.

    Parameters
    ----------
    signs : array_like
        A 1-d endogenous response variable. The dependent variable.
    """

    def __init__(self, signs: ArrayLike1D) -> None:
        self._signs = signs
        self._nb_signs = len(signs)

        # Will be set in fit()
        self._order = None
        
        self._x = None
        self._y = None
        self._first_order_signs = None
        self._start_idx_parameters = None

        # Will be set in simulate()
        self._simulation = None

    @abstractmethod
    def resid(self) -> np.ndarray:
        """
        Estimate and return the residuals of the model.
        """
        pass

    @abstractmethod
    def fit(self, method: str) -> TimeSeries:
        """
        Estimate the model parameters.
        """
        pass

    @abstractmethod
    def simulate(self, size: int) -> np.ndarray:
        """
        Simulate a time series of signs after the model has been fitted.
        """
        pass

    def calculate_acf(self, nb_lags: int, time_series: Optional[ArrayLike1D] = None) -> np.ndarray:
        """
        Calculate the autocorrelation function of a time series of signs.

        Parameters
        ----------
        nb_lags : int
            Number of lags to return autocorrelation for.
        time_series : array_like, default None
            The time series for which to compute the acf. If None, the original time series of the model is used.

        Returns
        -------
        np.ndarray
            The autocorrelation for lags 0, 1, ..., nlags.
            It includes the lag 0 autocorrelation (i.e., 1), thus the size is (nlags + 1,).
        """
        if time_series is None:
            time_series = self._signs

        check_condition(condition=nb_lags is not None and 1 <= nb_lags < len(time_series),
                        exception=IllegalNbLagsException(f"Can only calculate the autocorrelation function with a number of lags positive and lower than the time series length (requested number of lags {nb_lags} should be < {len(time_series)})."))
        return acf(x=time_series, nlags=nb_lags, qstat=False, fft=True, alpha=None, bartlett_confint=True, missing="raise")

    def calculate_pacf(self, nb_lags: int, alpha: Optional[float] = None, time_series: Optional[ArrayLike1D] = None) -> np.ndarray | Tuple[np.ndarray, np.ndarray]:
        """
        Calculate the partial autocorrelation function of a time series of signs.

        Parameters
        ----------
        nb_lags : int
            Number of lags to return autocorrelation for.
        alpha : float, optional
            If a number is given, the confidence intervals for the given level are returned.
            For example, if alpha=0.05, 95 % confidence intervals are returned.
        time_series : array_like, default None
            The time series for which to compute the pacf. If None, the original time series of the model is used.

        Returns
        -------
        pacf : np.ndarray
            The partial autocorrelation for lags 0, 1, ..., nb_lags.
            It includes the lag 0 autocorrelation (i.e., 1), thus the size is (nlags + 1,).
        confint : ndarray, optional
            Confidence intervals for the pacf at lags 0, 1, ..., nb_lags.
            The shape is (nlags + 1, 2). It is Returned if alpha is not None.
        """
        if time_series is None:
            time_series = self._signs

        check_condition(condition=1 <= nb_lags < len(time_series) // 2,
                        exception=IllegalNbLagsException(f"Can only calculate the partial autocorrelation function with a number of lags positive and lower than 50% of the time series length (requested number of lags {nb_lags} should be < {len(time_series) // 2})."))
        check_condition(condition=alpha is None or 0 < alpha <= 1,
                        exception=IllegalValueException(f"Alpha {alpha} is invalid, it must be in the interval [0, 1]"))
        return pacf(x=time_series, nlags=nb_lags, method="burg", alpha=alpha)

    def simulation_summary(self, plot: bool = True, log_scale: bool = True, percentiles: Tuple[float, ...] = (50.0, 75.0, 95.0, 99.0, 99.9)) -> pd.DataFrame:
        """
        Return a statistical summary comparing the original signs and the simulated ones.

        The statistics are computed over the series counting the number of consecutive signs in a row.

        The function is to be called after a model has been fitted and simulated.

        Parameters
        ----------
        plot : bool
            If True, plots two graphs. One comparing the autocorrelation function
            of the original and simulated time series, and another comparing the partial autocorrelation.
        log_scale : bool, default true
            If True, use a log scale for plotting graphs, otherwise use a linear scale.
            It has no effect if `plot` is False.
        percentiles : tuple of float
            The percentiles to use.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing the statistics for the original and simulated time series.
        """
        plot = bool_like(value=plot, name="plot", optional=False, strict=True)
        log_scale = bool_like(value=log_scale, name="log_scale", optional=False, strict=True)
        check_condition(self._simulation is not None, ModelNotSimulatedException("The model has not yet been simulated. Simulate the model first by calling 'simulate()'."))

        statistics_training = self._compute_signs_statistics(signs=self._signs, column_name="Training", percentiles=percentiles)
        statistics_simulation = self._compute_signs_statistics(signs=self._simulation, column_name="Simulation", percentiles=percentiles)
        statistics = pd.concat([statistics_training, statistics_simulation], axis=1).round(decimals=2)

        if plot:
            self._build_fig_corr_training_vs_simulation(log_scale=log_scale)

        return statistics

    @staticmethod
    def is_time_series_stationary(time_series: ArrayLike1D, nb_lags: Optional[int] = None, significance_level: float = 0.05, regression: Literal["c", "ct", "ctt", "n"] = "c") -> bool:
        """
        Test whether a time series is stationary at a given significance level using the Augmented Dickey-Fuller test.

        Parameters
        ----------
        time_series : array_like
            The time series to test for stationarity.
        nb_lags : int, default None
            The number of lags to include in the test. If None, the default used by the test is applied.
        significance_level : float, default 0.05
            The significance level for the test. If the p-value is less than or equal to this value, the time series is considered stationary.
        regression: {'c', 'ct', 'ctt', 'n'}, default 'c'
            Constant and trend order to include in regression.

            * "c" : constant only (default).
            * "ct" : constant and trend.
            * "ctt" : constant, and linear and quadratic trend.
            * "n" : no constant, no trend.

        Returns
        -------
        bool
            True if the time series is stationary at the given significance level, False otherwise.
        """
        df_test = stattools.adfuller(x=time_series, maxlag=nb_lags, regression=regression, autolag=None)
        p_value = df_test[1]

        is_stationary = p_value <= significance_level
        logger.info(f"The time series of signs is {'non-' if not is_stationary else ''}stationary at the significance level {significance_level} (p-value: {np.round(p_value, decimals=6)}, number of lags used: {df_test[2]}).")
        return is_stationary

    @staticmethod
    def breusch_godfrey_test(resid: np.ndarray, nb_lags: Optional[int] = None) -> Tuple[float, float]:
        """
        Perform the Breusch-Godfrey test for residual autocorrelation.

        Parameters
        ----------
        resid : np.ndarray
            The residuals from a regression model. Must be a 1-dimensional array.
        nb_lags : int, default None
            The number of lags to include in the test. If None, defaults to min(10, len(resid) // 5).

        Returns
        -------
        lagrange_multiplier : float
            The value of the Lagrange Multiplier test statistic.
        p_value : float
            The p-value for the test statistic.
        """
        resid = np.asarray(resid)
        if resid.ndim != 1:
            raise ValueError("Residuals must be a 1d array.")

        nb_resid = resid.shape[0]
        if nb_lags is None:
            nb_lags = min(10, resid.shape[0] // 5)

        x = lagmat(x=resid, maxlag=nb_lags, trim="forward", original="ex")
        x_with_cst = np.c_[np.ones(shape=nb_resid), x]

        res = OLS(resid, x_with_cst).fit()

        lagrange_multiplier = nb_resid * res.rsquared
        p_value = stats.chi2.sf(lagrange_multiplier, nb_lags)

        return lagrange_multiplier, p_value

    @classmethod
    def _compute_signs_statistics(cls, signs: ArrayLike1D, column_name: str, percentiles: Tuple[float, ...]) -> pd.DataFrame:
        series_nb_consecutive_signs = cls._compute_series_nb_consecutive_signs(signs=signs)
        names, values = [], []
        names.append("size"), values.append(len(signs))
        names.append("pct_buy (%)"), values.append(cls.proportion_buy(signs=signs))
        names.append("mean_nb_consecutive_values",), values.append(np.mean(series_nb_consecutive_signs))
        names.append("std_nb_consecutive_values"), values.append(np.std(series_nb_consecutive_signs))
        names.extend([f"Q{percentile}_nb_consecutive_values" for percentile in percentiles])
        values.extend(np.percentile(series_nb_consecutive_signs, percentiles))

        return pd.DataFrame(data=values, columns=[column_name], index=names)

    @staticmethod
    def _compute_series_nb_consecutive_signs(signs: ArrayLike1D) -> np.ndarray:
        series_nb_consecutive_signs = []
        current_nb = 1
        for i in range(1, len(signs)):
            if signs[i] == signs[i - 1]:
                current_nb += 1
            else:
                series_nb_consecutive_signs.append(current_nb)
                current_nb = 1

        series_nb_consecutive_signs.append(current_nb)
        assert np.sum(series_nb_consecutive_signs) == len(signs)
        return np.array(series_nb_consecutive_signs)

    @staticmethod
    def proportion_buy(signs: ArrayLike1D) -> float:
        return round(100 * sum([1 for sign in signs if sign == 1]) / len(signs), 2)

    def _build_fig_corr_training_vs_simulation(self, log_scale: bool = True) -> Figure:
        nb_lags = min(2 * self._order, len(self._signs) // 2 - 1)
        acf_training = self.calculate_acf(nb_lags=nb_lags)
        acf_simulation = self.calculate_acf(nb_lags=nb_lags, time_series=self._simulation)
        pacf_training = self.calculate_pacf(nb_lags=nb_lags, alpha=None)
        pacf_simulation = self.calculate_pacf(nb_lags=nb_lags, alpha=None, time_series=self._simulation)

        fig, axe = plt.subplots(1, 2, figsize=(16, 4))

        acf_title = f"ACF function for training and simulated time series"
        self._fill_axe(axe=axe[0], functions=[acf_training, acf_simulation], colors=["green", "purple"], linestyles=["dashed", "solid"], labels=["Training", "Simulation"], title=acf_title, xlabel="Lag", log_scale=log_scale, order=self._order)

        pacf_title = f"PACF function for training and simulated time series"
        self._fill_axe(axe=axe[1], functions=[pacf_training, pacf_simulation], colors=["green", "purple"], linestyles=["dashed", "solid"], labels=["Training", "Simulation"], title=pacf_title, xlabel="Lag", log_scale=log_scale, order=self._order)

        return fig

    @staticmethod
    def _fill_axe(axe: Any, functions: List[np.ndarray], colors: List[str], linestyles: List[str], labels: List[str], title: str, xlabel: str, log_scale: bool, order: Optional[int] = None) -> None:
        all_values = np.concatenate(functions)
        y_scale = f"{'log' if log_scale else 'linear'}"

        for function, color, linestyle, label in zip(functions, colors, linestyles, labels):
            axe.plot(function, color=color, linestyle=linestyle, label=label)

        axe.set_title(f"{title} ({y_scale} scale)")
        axe.set_yscale(y_scale)

        axe.set_xlabel(xlabel)
        axe.set_xlim(-1, max(len(function) for function in functions) - 1)

        y_min = max(0.0001, np.min(all_values)) if log_scale else np.min(all_values)
        axe.set_ylim(y_min, np.max(all_values) + 0.1)

        if order is not None:
            axe.axvline(x=order, color="blue", label="Order of the model", linestyle="--")

        axe.grid()
        axe.legend()

    def plot_acf_and_pacf(self, nb_lags: int, log_scale: bool = True, time_series: Optional[ArrayLike1D] = None) -> Figure:
        if time_series is None:
            time_series = self._signs

        acf_function = self.calculate_acf(nb_lags=nb_lags, time_series=time_series)
        pacf_function = self.calculate_pacf(nb_lags=nb_lags, alpha=None, time_series=time_series)

        fig, axe = plt.subplots(1, 2, figsize=(16, 4))
        self._fill_axe(axe=axe[0], functions=[acf_function], colors=["green"], linestyles=["solid"], labels=[f"Time series of size {len(time_series)}"], title="ACF function", xlabel="Lag", log_scale=log_scale, order=None)
        self._fill_axe(axe=axe[1], functions=[pacf_function], colors=["orange"], linestyles=["solid"], labels=[f"Time series of size {len(time_series)}"], title="PACF function", xlabel="Lag", log_scale=log_scale, order=None)

        return fig



class EnumValueException(Exception):
    """Raised when the enum value is not valid"""
    pass


class UnsupportedOsException(Exception):
    """Raised if the OS is not supported when loading a shared library"""
    pass


class SharedLibraryNotFoundException(Exception):
    """Raised if the shared library is not found in the registry"""
    pass
